// set the dimensions and margins of the graph
const width = 1200;
const height = 800;
const mapWidth = width * 0.55;
const plotWidth = width - mapWidth + 200;
const margin = 60;

// Let's create two svg elements and append them to my top-level div
// Here's the first svg, and I'll save the selection to the variable "svg"
const svg = d3
  .select('#my_dataviz')
  .append('svg')
  .attr('width', mapWidth)
  .attr('height', height);

svg
  .append('text')
  .attr('x', mapWidth / 2)
  .attr('y', 225)
  .attr('text-anchor', 'middle')
  .style('font-size', '20px')
  .text(
    'World map based on Total U.S. Military aid since 1946',
  );

// And the second; the ".append" method returns a selection with just
// the svg in it, so I can manipulate it later (by appending or joining
// data with it)
const svg2 = d3
  .select('#my_dataviz')
  .append('svg')
  .attr('width', plotWidth)
  .attr('height', height);

const tooltip = d3
  .select('#my_dataviz')
  .append('div')
  .attr('id', 'tooltip');

// Here are the names of the attributes I'll pull out of my data
// objects for my various visual channels
// It's useful to declare them once, since I will need to refer to them
// in several places (e.g. setting visual properties, axes, and labels)

const mapColorAttr = 'current_amount';
const clickedCountries = new Set();
const plotColorAttr = 'Continent';
const xAttr = 'Fiscal Year';
const yAttr = 'constant_amount';
const rAttr = 'constant_amount';

var xExtent = undefined;
var yExtent = undefined;
var xScale = d3.scaleLinear();
var yScale = d3.scaleLinear();
var xScaleType = 'linear';
var yScaleType = 'linear';

// Here's a snippet of code that I ended up using in a couple places,
// so I've refactored it out as a function
//     data: an iterable (typically Array) of all data points
//     attr: the name of the attribute you want the extent of
//     returns: [min, max] of the attr values in data
// Note: min and max only really makes sense for numeric types, so I've
// included a typecast (the "+") - this will keep me from accidentally
// getting tripped up when my attribute is really strings representing
// numbers
function getDataExtent(data, attr) {
  return d3.extent(
    // For my data, missing values show up as empty strings, which
    // JavaScript then casts to 0. I will filter out all of these to make
    // sure I get the true extent of the non-missing data
    d3.map(data, (r) => +r[attr]).filter((r) => r > 0),
  );
}

// Here's the main rendering code; whenever I call this function, it
// will draw a choropleth map from the mapData and a scatter plot from
// the countryData
function render(mapData, countryData) {
  const israelData = countryData.get('israel');

  // Here's how I'm going to convert lat/long coordinates into pixel values
  // on the screen; I can use whichever projection I like best
  var projection = d3
    .geoEqualEarth()
    .fitSize([mapWidth, height], mapData);
  // And here's the part that converts a sequence of points around a
  // polygon into a valid "d" attribute for a <path> tag
  var path = d3.geoPath(projection);

  // Update the domain of the scale with the new colorExtent
  var mapColorScale = d3
    .scaleThreshold()
    .domain([30000000000, 60000000000, 90000000000])
    .range(['#bdc9e1', '#67a9cf', '#02818a', '#016c59']);

  const legend = svg
    .append('g')
    .attr('class', 'legend')
    .attr('transform', 'translate(50,600)');
  //.attr('y', 225);

  const colorRange = mapColorScale.range();
  const domainValues = mapColorScale.domain();

  const legendItem = legend
    .selectAll('.legend-item')
    .data(colorRange)
    .enter()
    .append('g')
    .attr('class', 'legend-item')
    .attr('transform', (d, i) => `translate(0, ${i * 20})`);

  legendItem
    .append('rect')
    .attr('width', 18)
    .attr('height', 18)
    .style('fill', (d) => d);

  legendItem
    .append('text')
    .attr('x', 24)
    .attr('y', 9)
    .attr('dy', '.35em')
    .text((d, i) => {
      if (i === 0) {
        return '< ' + domainValues[i].toLocaleString();
      } else if (i === colorRange.length - 1) {
        return '≥ ' + domainValues[i - 1].toLocaleString();
      } else {
        return (
          domainValues[i - 1].toLocaleString() +
          ' - ' +
          domainValues[i].toLocaleString()
        );
      }
    });

  // Update the domain of the color scale to match the extent of summed values
  // Here are the rest of the scaling functions (for the scatter plot)
  // We'll need to handle the x/y positions, radius, and color

  // Boring linear scale for x
  xExtent = getDataExtent(israelData, xAttr);
  xScale = d3
    .scaleLinear()
    .domain(xExtent)
    .range([margin, plotWidth - margin]);

  // Boring linear scale for y
  yExtent = getDataExtent(israelData, yAttr);
  yScale = d3
    .scaleLinear()
    .domain(yExtent)
    // Have you noticed that the range for the y axis is backwards?
    // It maps the lowest input to the highest pixel value, and the highest
    // input to 0 pixels. This is because the origin (0,0) for graphics is
    // in the top left, instead of the bottom left like mathematicians might
    // expect
    .range([height - margin, margin]);

  const rScale = d3
    .scaleLinear()
    .domain(getDataExtent(israelData, rAttr))
    .range([3, 10]);

  // Since we need to do a little error checking here, lets write a
  // nice function so that our attribute setting looks cleaner
  function getCountryColor(d) {
    // First, try to get the data for a given country
    let entry = countryData.get(d.properties.name);
    if (entry === undefined) {
      return 'grey';
    } else {
      let value = sumConstantAmount(entry);
      // Scale the value to determine the color
      return mapColorScale(value);
    }
  }

  // This one is way more straightforward, since I assume I always have
  // complete data for each point on the scatter plot
  function getPointColor(d) {
    let entry = d['Country Name'];
    if (entry === undefined) {
      return 'grey';
    } else {
      let test = intToRGB(hashCode(entry));
      return test;
    }
  }

  function hashCode(str) {
    // java String#hashCode
    var hash = 0;
    for (var i = 0; i < str.length; i++) {
      hash = str.charCodeAt(i) + ((hash << 5) - hash);
    }
    return hash;
  }

  function intToRGB(i) {
    var r = Math.floor(Math.sin(i) * 128 + 128);
    var g = Math.floor(Math.sin(i + 2) * 128 + 128);
    var b = Math.floor(Math.sin(i + 4) * 128 + 128);

    return '#' + rgbToHex(r, g, b);
  }

  function rgbToHex(r, g, b) {
    return (
      componentToHex(r) +
      componentToHex(g) +
      componentToHex(b)
    );
  }

  function componentToHex(c) {
    var hex = c.toString(16);
    return hex.length == 1 ? '0' + hex : hex;
  }

  // Ok, down to brass tacks: let's add <path> elements to our first svg
  // that encode the borders of each country. I'll save this selection as
  // a variable, so that I can modify the countries later if I need to
  const countryMarks = svg
    // I like to append a <g> element that will hold all my paths, just so
    // I have the option of easily adjusting or moving the entire group
    .append('g')
    // Since the svg is originally empty (I just created it, after all), any
    // selection here will be empty; its only purpose is to give me a place
    // to join my data
    .selectAll('path')
    // Once I have the (empty) selection, I will join all of my data points to
    // it. I know that the structure of geojson is for there to be a "features"
    // attribute that contains the Array of all the geo objects, and I want to
    // bind each geo object to its own path
    .data(mapData.features)
    .join('path')
    // Ok, now that I have a shiny new <path> element with a single country's
    // geo data bound to it, I'm ready to start setting them up.
    // First, the fill color: I'll use the function I wrote above that converts
    // a country name string into a scaled color value
    .attr('fill', getCountryColor)
    // Next, I'll set the "id" attribute, which will give me the ability to
    // select linked items that have the same id
    .attr('id', (d) => d.properties.name)
    .classed('poly', true)
    // Finally, lets use the default geojson setup of setting the 'd' attribute
    // using our path generator function. This is sort of magic that works under
    // the hood
    .attr('d', path);

  svg2
    .append('text')
    .attr('x', plotWidth / 2)
    .attr('y', margin - 15)
    .attr('text-anchor', 'middle')
    .style('font-size', '20px')
    .text(
      'U.S. Foreign Aid overtime in comparison to Israel',
    );
  svg2
    .append('g')
    .attr('transform', `translate(0,${height - margin})`) // This controls the vertical position of the Axis
    .call(d3.axisBottom(xScale).tickFormat(d3.format('d'))); // Use d3.format('d') to format ticks as integers without commas

  svg2
    .append('text')
    .text(xAttr)
    .attr('x', plotWidth / 2)
    .attr('y', height - margin / 2)
    .style('text-anchor', 'middle');

  svg2
    .append('g')
    .attr('id', 'yAxisG')
    .attr('transform', `translate(${margin},0)`) // This controls the vertical position of the Axis
    .call(d3.axisRight(yScale));

  svg2
    .append('text')
    .text('Aid in USD Accounting for Inflating')
    .attr('transform', 'rotate(-90)')
    .attr('x', -height / 1.5)
    .attr('y', margin / 2);

  // Now we need to add points to our scatter plot; again I will save the
  // selection as a variable so I can modify it more later
  const scatterMarks = drawScatterPlotMarks(israelData);
  addScatterPlotEventListeners(scatterMarks);
  // Okay! It's later now: let's hook up the mouse events for our countries
  // and points in the scatter plot
  function drawScatterPlotMarks(data) {
    console.log(data);
    return svg2
      .append('g')
      .selectAll('circle')
      .data(data)
      .join('circle')
      .attr('r', (d) => rScale(d[rAttr]))
      .attr('fill', getPointColor)
      .attr(
        'id',
        (d) =>
          `${normalizeCountryName(d['Country Name'])}-${d[xAttr]}`,
      )
      .attr('class', (d) => `dot ${d.Continent}`)
      .style('opacity', 0.4)
      .attr('cx', (d) => xScale(d[xAttr]))
      .attr('cy', (d) => yScale(d[yAttr]));
  }
  function arraysEqual(arr1, arr2) {
    return arr1.every((element) => arr2.includes(element));
  }
  // Planning ahead, the first thing I'm going to want to do is highlight
  // matching marks in the linked plots, and I know they're linked by their
  // "id" attribute matching, and differentiated by having different classes
  // ("poly" for the map marks, and "dot" for the scatterplot marks)
  function highlightMarks(id) {
    d3.selectAll(
      `circle[id="${normalizeCountryName(id['Country Name'])}-${id['Fiscal Year']}"]`,
    )
      .attr('r', 8)
      .style('opacity', 1)
      .style('stroke', 'black');
  }

  // Finally, I can use my scaling functions for color and radius to reset
  // back to normal when I mouse out
  function resetMarks(id) {
    d3.selectAll(
      `circle[id="${normalizeCountryName(id['Country Name'])}-${id['Fiscal Year']}"]`,
    )
      .attr('fill', getPointColor)
      .attr('r', (d) => rScale(d[rAttr]))
      .style('opacity', 0.4)
      .style('stroke', 'none');
  }

  function popUpTooltip(event, data) {
    //console.log(data);
    tooltip.style('opacity', 1);
    tooltip.html(
      `${data['Country Name']}<br/>Year: ${data['Fiscal Year']} <br/> Amount: $${parseFloat(data['constant_amount']).toLocaleString()}`,
    );
  }

  function MappopUpTooltip(event, data, y) {
    tooltip.style('opacity', 1);
    tooltip.html(
      `${data.properties.name}<br/>Total Aid: ${y}`,
    );
  }

  function MapadjustTooltip(event) {
    tooltip
      .style('left', event.x - 80 + 'px')
      .style('top', event.y - 80 + 'px');
  }

  function adjustTooltip(event) {
    tooltip
      .style('left', event.x - 80 + 'px')
      .style('top', event.y - 80 + 'px');
  }

  function sumConstantAmount(data) {
    let totsum = 0;
    data.forEach((entry) => {
      totsum += parseFloat(entry.constant_amount);
    });
    return totsum.toExponential(2);
  }

  // With these behaviors captured in functions, we just need to hook them up
  // as event listeners on our mark selections!
  countryMarks
    .on('mouseover', function (event, data) {
      let x = countryData.get(data.properties.name);
      let y = sumConstantAmount(x);
      MappopUpTooltip(event, data, y);
    })
    .on('mousemove', function (event, data) {
      MapadjustTooltip(event);
    })
    .on('mouseleave', function (event, data) {
      tooltip.style('opacity', 0);
    })
    .on('click', function (event, data) {
      const countryName = data.properties.name;
      let col = countryData.get(countryName);
      let col2 = col[0];
      if (clickedCountries.has(countryName)) {
        svg2
          .selectAll(
            `circle[id^="${normalizeCountryName(countryName)}"]`,
          )
          .remove();
        d3.select(`#${countryName}`).attr(
          'fill',
          getCountryColor,
        );
        clickedCountries.delete(countryName);
      } else {
        const clickedCountryData =
          countryData.get(countryName);
        const scatterMarks = drawScatterPlotMarks(
          clickedCountryData,
        );
        addScatterPlotEventListeners(scatterMarks);
        d3.select(`#${countryName}`).attr(
          'fill',
          getPointColor(col2),
        );
        clickedCountries.add(countryName);
      }
    });

  function addScatterPlotEventListeners(scatterMarks) {
    scatterMarks
      .on('mouseover', function (event, data) {
        popUpTooltip(event, data);
        highlightMarks(data);
      })
      .on('mousemove', function (event, data) {
        adjustTooltip(event);
      })
      .on('mouseleave', function (event, data) {
        resetMarks(data);
        tooltip.style('opacity', 0);
      });
  }

  d3.select('#yAxisG').on('click', (event, data) =>
    toggleAxisScale(),
  );
}

function toggleAxisScale() {
  if (yScaleType == 'linear') {
    yScale = d3
      .scaleLog()
      .domain(yExtent)
      .range([height - margin, margin]);
    yScaleType = 'log';
  } else if (yScaleType == 'log') {
    yScale = d3
      .scaleLinear()
      .domain(yExtent)
      .range([height - margin, margin]);
    yScaleType = 'linear';
  }

  svg2.select('#yAxisG').remove();

  svg2
    .append('g')
    .attr('id', 'yAxisG')
    .on('click', () => toggleAxisScale())
    .attr('transform', `translate(${margin},0)`) // This controls the vertical position of the Axis
    .transition()
    .duration(2000)
    .call(d3.axisRight(yScale));

  svg2
    .selectAll('circle.dot')
    .transition()
    .duration(2000)
    .attr('cy', (d) => yScale(d[yAttr]));
}

//Read the data
const mapUrl =
  'https://raw.githubusercontent.com/holtzy/D3-graph-gallery/master/DATA/world.geojson';
const countryStatsUrl =
  'https://raw.githubusercontent.com/thehankstah/cp341/main/US_foreign_aid_country_final.csv';
const continentUrl =
  'https://raw.githubusercontent.com/samayo/country-json/master/src/country-by-continent.json';

function normalizeCountryName(s) {
  return s.replaceAll(' ', '_').toLowerCase();
}

function normalizeYears(s) {
  return s.replaceAll(',', '');
}

function parseNumStr(s) {
  return +s.replaceAll(',', '').replaceAll('$$', '');
}

Promise.all([
  d3.json(mapUrl),
  d3.csv(countryStatsUrl),
  d3.json(continentUrl),
]).then(([mapData, csvData, continentData]) => {
  var countryData = new Map();
  var countryToMapNames = new Map([
    ['United States', 'USA'],
    ['Tanzania', 'United Republic of Tanzania'],
    ['United Kingdom', 'England'],
    ['United Kingdom', 'Wales'],
    ['Republic of Ireland', 'Ireland'],
    ['China (P.R.C.)', 'China'],
    ['Korea, Democratic Republic of', 'South Korea'],
  ]);
  csvData.forEach((row) => {
    if (countryToMapNames.has(row['Country Name'])) {
      row['Country Name'] = countryToMapNames.get(
        row['Country Name'],
      );
    }

    row.Name = normalizeCountryName(row['Country Name']);
    row['Fiscal Year'] = normalizeYears(row['Fiscal Year']);
    row.constant_amount = parseNumStr(
      row['constant_amount'],
    );
    if (countryData.has(row.Name)) {
      const existingEntry = countryData
        .get(row.Name)
        .find(
          (entry) =>
            entry['Fiscal Year'] === row['Fiscal Year'],
        );
      if (existingEntry) {
        existingEntry.constant_amount +=
          row.constant_amount;
      } else {
        countryData.get(row.Name).push(row);
      }
    } else {
      countryData.set(row.Name, [row]);
    }
  });
  mapData.features.forEach((f) => {
    f.properties.name = normalizeCountryName(
      f.properties.name,
    );
  });

  console.log(mapData);
  render(mapData, countryData);
})